#include "testlib.h"
#include <iostream>
#include <set>
#include <vector>
#include <cassert>

using namespace std;
/**
    TO DO: more back edges
    check what happens if an edges doesn't have an scc
**/
const int LOGMAX = 22;
struct SccGraph{
    int n, m, totalCompSize, compCount;
    set< pair<int, int> > edges;
    vector<int> comp;
    SccGraph(int _n, int _m){
        n = _n;
        m = _m;
        totalCompSize = 0;
        compCount = 0;
        comp.resize(n+1);
    }
    void addEdge(int a, int b){
        edges.insert({a,b});
        if(edges.size() > m){
            throw "Too many edges";
        }
    }
    void clique(int l, int r){
        for(int i=l;i<=r;i++)
            for(int j=i+1;j<=r;j++){
                addEdge(i, j);
                addEdge(j, i);
            }
    }
    void cycle(int l, int r){
        for(int i=l;i<r;i++)
            addEdge(i, i+1);
        if(l <= r) {
          addEdge(r, l);
        }
    }
    void cycleUnion(int l, int r){
        int node = l;
        while(node != r){
            int sz = rnd.next(1, r - node + 1);
            if(sz == 1) continue;
            cycle(node, node+sz-1);
            node = node + sz - 1;
        }
    }
    void randomBackwardsEdges(int l, int r, int freq=5){
        int reached = l;
        for(int node=l; node<=r; node++){
            if(node<r) addEdge(node, node+1);
            if(reached < node && (rnd.next(freq)==0||node==r)){
                addEdge(node, rnd.next(l, reached));
                reached = l;
            }
        }
    }
    void connectSccsLine(){
        for(int i=1;i<n;i++)
            if(comp[i]!=comp[i+1])
                addEdge(i, i+1);
    }
    void connectSccsRandomDag(){
        int left = m - edges.size();
        while(left > 0){
            int a = rnd.next(1, n);
            int b = rnd.next(1, n);
            if(a>b) swap(a,b);
            if(comp[a]==comp[b]) continue;
            addEdge(a, b);
            left--;
        }
    }
    void addRandomEdges(int howMany){
        int left = howMany;
        while(left > 0){
            int a = rnd.next(1, n);
            int b = rnd.next(1, n);
            if(a>b) swap(a,b);
            if(a==b) continue;
            addEdge(a, b);
            left--;
        }
    }
    void addComp(int sz, int type=0){
        assert(totalCompSize + sz <= n);
        compCount++;
        for(int i=totalCompSize + 1; i <= totalCompSize + sz; i++){
            comp[i] = compCount;
        }
        int l = totalCompSize + 1, r = totalCompSize + sz;
        if(type == 0){
            clique(l, r);
        }
        if(type == 1){
            cycle(l, r);
        }
        if(type == 2){
            cycleUnion(l, r);
        }
        if(type == 3){
            randomBackwardsEdges(l, r);
        }
        totalCompSize += sz;
    }
    void addClique(int sz){
        addComp(sz, 0);
    }
    void addCycle(int sz){
        addComp(sz, 1);
    }
    void addCycleUnion(int sz){
        addComp(sz, 2);
    }
    void addRandomBackwardsEdges(int sz){
        addComp(sz, 3);
    }
    vector<pair<int,int>> getShuffledEdges(){
        vector<int> ord(n);
        for(int i=0;i<n;i++){
            ord[i] = i + 1;
        }
        shuffle(ord.begin(),ord.end());
        vector<pair<int,int>> finalEdges;
        for(auto edg: edges){
            finalEdges.push_back({ord[edg.first-1], ord[edg.second-1]});
        }
        shuffle(finalEdges.begin(), finalEdges.end());
        return finalEdges;
    }

};
struct Test{
    int type, n, q;
    vector<pair<int,int>>edges;
    Test(int _type, int _n, int _q, vector<pair<int,int>> _edges){
        type = _type;
        n = _n;
        q = _q;
        edges = _edges;
    }
};
Test logarithmicGraph(int type, int n, int m, int q){
    SccGraph graph(n, m);
    int cnt = 0 ;
    for(int p=LOGMAX;p>=0;p--){
        if(((1<<p)&n)){
            graph.addComp((1<<p), (++cnt)%3+1);
        }
    }
    graph.connectSccsLine();
    return Test(type, n, q, graph.getShuffledEdges());
}
Test smallCliques(int type, int n, int m, int q, int cliqueSize=4){
    SccGraph graph(n, m);
    int sz = n;
    while(sz > 0){
        graph.addClique(min(sz,cliqueSize));
        sz -= cliqueSize;
    }
    graph.connectSccsRandomDag();
    return Test(type, n, q, graph.getShuffledEdges());
}
Test veryBigComponent(int type, int n, int m, int q, int bigComponent){
    assert(bigComponent <= n);
    SccGraph graph(n, m);
    graph.addRandomBackwardsEdges(bigComponent);
    if(n-bigComponent){
        graph.addCycle(n-bigComponent);
    }
    graph.connectSccsLine();
    return Test(type, n, q, graph.getShuffledEdges());
}
Test sqrtGraph(int type, int n, int m, int q){
    SccGraph graph(n, m);
    int cnt = 0 ;
    int sum = 0;
    for(int p=1;p<=n&&sum+p<=n;p++){
        graph.addComp(p, (++cnt)%3+1);
        sum += p;
    }
    graph.addCycle(n-sum);
    graph.connectSccsRandomDag();
    return Test(type, n, q, graph.getShuffledEdges());
}
Test randomGraph(int type, int n, int m, int q){
    SccGraph graph(n, m);
    graph.addRandomEdges(m);
    return Test(type, n, q, graph.getShuffledEdges());
}
void printTest(Test test){
    cout<<test.type<<'\n';
    cout<<test.n<<' '<<test.edges.size()<<' '<<test.q<<'\n';
    for(auto edge: test.edges)
        cout<<edge.first<<' '<<edge.second<<'\n';
}
int main(int argc, char** argv) {
    registerGen(argc, argv, 1);
    int type = opt<int>("type");
    int n = opt<int>("n");
    int m = opt<int>("m");
    int q = opt<int>("q");
    int smallConst = opt<int>("smallConst");
    int bigConst = opt<int>("bigConst");
    string mode = opt<string>("mode");
  
    if(mode == "logarithmicGraph") {
      printTest(logarithmicGraph(type, n, m, q));       
    } else if(mode == "veryBigComponent") {
      printTest(veryBigComponent(type, n, m, q, n/2));
    } else if(mode == "veryBigComponent2") {
      printTest(veryBigComponent(type, n, m, q, n-smallConst));
    } else if(mode == "smallCliques") {
      printTest(smallCliques(type, min(m/smallConst, n), m, q, smallConst));
    } else if(mode == "sqrtGraph") {
      printTest(sqrtGraph(type, n, m, q));
    } else if(mode == "smallCliques2") {
      int smallN = min(n, m/bigConst/smallConst);
      printTest(smallCliques(type, smallN, m, q, bigConst));
    } else if(mode == "randomGraph") {
      printTest(randomGraph(type, n/smallConst, m, q));
    } else if(mode == "randomGraph2") {
      printTest(randomGraph(type, min(n, 2*n/smallConst), m, q));
    } else {
      assert(false);
    }

    return 0;
}
